## Score computer for the EvAAL competition
##
## Copyright © Francesco Potortì, 2015-2016
## Licensed to anyone under the GNU Affero General Public License 3.0

## $Revision: 2.35 $

## When this script is called from Octave it looks under BASEDIR for
## directories containing all the necessary files (see below) and shows
## them in a menu for the user to choose from.  After the choice is
## done, data are read and a report is created.  If doplot is true
## a graph displaying the estimated path superimposed over the map
## is displayed and saved under the chosen dir in a pdf file.

## The normal way to call this script from the Octave prompt is:
## >> evaalscore
## or
## >> doplot=false; basedir="../competitors/data"; evaalscore
## where doplot defaults to true and basedir defaults to "."

## If you want to completely avoid the prompt asking to choose a directory
## you can set the directory name beforehand like this:
## >> doplot=false; basedir="../competitors/data"; dirname="C1"; evaalscore

## If you want the plot but avoid the overhead of creating the path file:
## >> dopath=false; basedir="../competitors/data"; dirname="C1"; evaalscore

## The necessary files are 'buttonsPressed.log', 'positions.log', which
## are created by the stepLogger app, and the file 'distance', which
## contains one line of text: either "localdistance" or "wgs84distance",
## depending on the reference system used for the coordinates.  It must
## be either a local system with unit in meters or else WGS84. In the
## latter case, the coordinates must be in the order long, lat.

## If files relative to a defined scenario are present, the relative
## report is created.  Maps should be PNG 8 bit image files used as the
## graph backgrounds, '*.corners' should be text files containing the
## ref system coordinates of the four corners of the map in the usual
## image order: NO, NE, SE, SW.


## Scenarios are tried in order starting from the first one.

scenarios = struct();

################ IPIN 2017 scenario
s = struct();
s.reporttitle = "IPIN competition 2017, Sapporo (JP)";
s.floorpenalty = 15;	# Floor penalty [m] used when computing distance

## Maps info
s.spsr = 2;				# subplot rows
s.spsc = 3;				# subplot columns
s.spi_legax = 0;			# subplot index of legend axes (0 means use file)
s.spi_legend = s.spsc;			# subplot index for legend and stats
s.spp_legend = [0.75 0.75 0.20 0.20];	# legend position
s.spi_distrib = s.spsr*s.spsc;		# subplot index for distribution graph
s.spp_distrib = [0.75 0.05 0.20 0.20];	# distribution graph position
s.statx = -4;				# statistics xpos relative to distr. graph
s.staty = 2.8;				# statistics ypos relative to distr. graph
s.mapinfo = struct("map", {"P1-HOK", "P0-HOK" # file basename
			  },
		   "rm", {0, 1},	# index of reference map for scaling
		   "title", {"First floor", "Ground floor"
			    },
		   "floor", {1, 0},
		   "spi", {1:2, 4:5	# subplot index
			  },
		   "spp", {[.15 .54 .45 .45], [.15 .00 .45 .45],
			  },
		   "psz", [],		# pixel width and height
		   "p2r", [],		# pixel to ref tform
		   "ah", []		# axes handle
		  );

scenarios.evaal2017 = s; clear s;

################ EvAAL-IPIN 2016 scenario
s = struct();
s.reporttitle = "EvAAL-IPIN competition 2016, Alcalá de Henares (ES)";
s.floorpenalty = 15;	# Floor penalty [m] used when computing distance

## Maps info
s.spsr = 2;				# subplot rows
s.spsc = 5;				# subplot columns
s.spi_legax = 0;			# subplot index of legend axes (0 means use file)
s.spi_legend = s.spsc;			# subplot index for legend
s.spp_legend = [0.84 0.75 0.20 0.20];	# legend position
s.spi_distrib = s.spsr*s.spsc;		# subplot index for distribution graph
s.spp_distrib = [0.84 0.05 0.20 0.20];	# distribution graph position
s.statx = -4;				# statistics xpos relative to distr. graph
s.staty = 2.8;				# statistics ypos relative to distr. graph
s.mapinfo = struct("map", { "P3_UAH",  "P2_UAH" # file basename
			    "P0_UAH",  "P1_UAH"
			  },
		   "rm", {0, 0; 0, 0},	# no map scaling
		   "title", {"Third floor",	"Second floor"
			     "Ground floor",	"First floor"
			    },
		   "floor", {3, 2; 0, 1},
		   "spi", {1:2, 3:4	# subplot index
			   6:7, 8:9
			  },
		   "spp", {[.00 .50 .37 .37], [.42 .50 .37 .37],
			   [.00 .00 .37 .37], [.42 .00 .37 .37],
			  },
		   "psz", [],		# pixel width and height
		   "p2r", [],		# pixel to ref tform
		   "ah", []		# axes handle
		  );

scenarios.evaal2016 = s; clear s;

################ EvAAL-ETRI 2015 scenario
s = struct();
s.reporttitle = "EvAAL-ETRI competition 2015, Banff (CA)";
s.floorpenalty = 15;	# Floor penalty [m] used when computing distance

## Maps info
s.spsr = 3;				# subplot rows
s.spsc = 4;				# subplot columns
s.spi_legax = 0;			# subplot index of legend axes (0 means use file)
s.spi_legend = s.spsc;			# subplot index for legend
s.spp_legend = [0.75 0.75 0.20 0.20];	# legend position
s.spi_distrib = s.spsr*s.spsc;		# subplot index for distribution graph
s.spp_distrib = [0.77 0.05 0.20 0.20];	# distribution graph position
s.statx = 2;				# statistics xpos relative to distr. graph
s.staty = 2;				# statistics ypos relative to distr. graph
s.mapinfo = struct("map", { "KCCI-L3",  "PDC-L3" # file basename
			    "KCCI-L2",  "PDC-L2"
			    "KCCI-L1",  "PDC-L1"
			  },
		   "rm", {0, 0; 0, 0; 0, 0},	# no map scaling
		   "title", {"maps not to scale",	""
			     "",			""
			     "",			""
			    },
		   "floor", {3, 3; 2, 2; 1, 1},
		   "spi", {1:2, 3		# subplot index
			   5:6, 7
			   9:10, 11
			  },
		   "spp", {[.00 .64 .45 .29], [.45 .64 .30 .29],
			   [.00 .32 .45 .29], [.45 .32 .30 .29],
			   [.00 .00 .45 .29], [.45 .00 .30 .29],
			  },
		   "psz", [],		# pixel width and height
		   "p2r", [],		# pixel to ref tform
		   "ah", []		# axes handle
		  );

scenarios.evaal2015 = s; clear s;

################ End of scenarios

#

## Work around some Octave 4.0 imread bugs
function img = fileimread(file)
  img = imread(file);
  if (size(img,3) == 3 && islogical(img))	# work around Octave bug #49140
    img = double(img);				# convert to double
#  elseif (isgray(img))				# work around Octave bug #49137
#    img = repmat(img, 1, 1, 3);		# convert to RGB
  endif
endfunction


## Draw the path only, without errors
function draw_path (scnr, directory)
  gtruthfile = [directory "/groundTruth.txt"];
  if !exist(gtruthfile, 'file')
    error("%s does not exist: cannot draw path\n", gtruthfile);
  endif

  buttonsfile = [directory "/buttonsPressed.log"];
  positionsfile = [directory "/positions.log"];
  pathfile = [directory "/path.pdf"];
  buttonsfilesaved = [buttonsfile ".saved"];
  positionsfilesaved = [positionsfile ".saved"];

  unwind_protect
    bfren = pfren = false;
    bfexist = exist(buttonsfile, 'file');
    pfexist = exist(positionsfile, 'file');
    if (bfexist) rename(buttonsfile, buttonsfilesaved); bfren = true; endif
    if (pfexist) rename(positionsfile, positionsfilesaved); pfren = true; endif

    ## Create a dummy buttonsPressed.log
    [l x y f] = textread(gtruthfile, "%s%f%f%f");
    bfh = fopen(buttonsfile, 'w');
    for t = 1:length(l)
      fprintf(bfh, "%ld : %s : 0 : 0 : 0\n", 1000*(fix(time())+t), l{t});
    endfor
    fclose(bfh);

    ## Create path-only output
    create_dummy_positions(directory, 0);
    scenariograph(scnr, directory, true, false);
    print(pathfile);

  unwind_protect_cleanup
    if (bfren) rename(buttonsfilesaved, buttonsfile); endif
    if (pfren) rename(positionsfilesaved, positionsfile); endif
  end_unwind_protect

endfunction


## Create a random "positions.log" file useful for testing
## N is how many new pos to create between existing ones
## iF n IS 0, create perfect positions, that is, without errors
##
## Call like this:
##  create_dummy_positions ("testevaal", 5)
function  create_dummy_positions (directory, n)
  positionsfile = [directory "/positions.log"];

  t = textread([directory "/buttonsPressed.log"], "%f%*%*%*%*", 'delimiter', ":");
  [l x y f] = textread([directory "/groundTruth.txt"], "%s%f%f%f");
  if (length(t) != length(l))
    warning("%d ground truth labels, but %d button pressures\n",
	  length(l), length(t));
    return
  endif

  if (n == 0)			# copy reference points
    spd = [t-1 x y f];
  else				# interpolate and perturbate ref points
    p = interp1([t x y f], 0:1/(n+1):length(t)+2, 'extrap');
    absdiff = quantile(abs(diff(p)), 0.5)
    d = n * (rand(size(p))-1) .* absdiff;
    pd = p+d;
    pd(:,4) = round(pd(:,4));		# floor must be integer
    spd = sortrows(pd);			# sort by increasing times
  endif

  fid = fopen(positionsfile, 'w');
  fprintf(fid, "%.0f\t%.6f\t%.6f\t%d\n", spd');
  fclose(fid);
endfunction

## Create the contents of an IMAGE.corners file from IMAGE.png and IMAGE.points
function mapcorners = create_corners(image, projection = 'affine', opt=0)
  pkg load image

  imagefile = [ image ".png"];
  pointsfile = [ image ".points"];
  cornersfile = [ image ".corners"];

  ## Create coordinate conversion matrix pixel_to_ref from pointsfile
  pw = imfinfo(imagefile).Width;
  ph = imfinfo(imagefile).Height;
  corners_pixel = [0,0; pw,0; pw,-ph; 0,-ph];
  [mapX mapY pixelX pixelY] = textread(pointsfile, "%f%f%f%f%*", 'delimiter', ",", 'headerlines', 1);
  pixel_to_ref = cp2tform([pixelX pixelY], [mapX mapY], projection, opt);

  ## Create image corners cordinates and store them into cornersfile
  mapcorners = tformfwd(pixel_to_ref, corners_pixel);
  if (nargout == 0)
    printf("Contents to be saved into %s:\n", cornersfile);
    printf("%12.6f%12.6f\n", mapcorners');
  endif

  ## cornersfile = [ image ".corners"];
  ## save_precision(10, 'local');
  ## save("-text", cornersfile, "mapcorners");

endfunction

#

## Look in BASEDIR for directories containing the necessary files and
## ask the user to choose one of them, return the chosen dir as a string,
## that is the dir name relative to BASEDIR
function dir = chosendir(basedir);

  ## The files expected in a dir with one or no maps.  Optional files "map"
  ## and "map_corners" are required to add a background map
  files_single_map = { "buttonsPressed.log",
		       "positions.log",
		       "distance"};

  files = readdir(basedir);
  gooddiridx = [];
  for i = 1:length(files)
    if (isdir([basedir "/" files{i}])
	&& !(strcmp(".", files{i}) || strcmp("..", files{i})))
      dirfiles = readdir([basedir "/" files{i}]);
      found_single = 0;
      for j = 1:length(dirfiles)
	found_single += any(strcmp(dirfiles{j}, files_single_map));
      endfor
      if (found_single == length(files_single_map))
	gooddiridx(end+1) = i;
      endif
    endif
  endfor
  dirs = files(gooddiridx);

  assert(length(dirs) > 0, "no good directories found under '%s', exiting\n", basedir);
  choice = 0;
  choice = menu("Choose a directory", {"no choice: EXIT" dirs{:}});
  if (choice < 1)
    dir = ""
  else
    dir = dirs{choice-1}	# first choice is "no choice"
  endif

endfunction


## Convert vectors of X and Y coordinates to pixel coords
function [px py] = to_pixel_coords (ph, pixel_to_ref, x, y)
  p = tforminv(pixel_to_ref, [x y]);
  px = p(:,1);
  # As usual, we set the image origin in the top-left corner but, like
  # QGIS' georeferencer, we do not follow the usual convention of y axis
  # going downwards, consequently y coordinates are negative.
  # To adjust things, below we use flipud on the image.
  py = p(:,2) + ph;
endfunction


## Distance in meters given longitude and latitude
## Points P1 and P2 are [long lat floor] or nx3 matrices
function d = wgs84distance(p1, p2, floorpenalty = 0)
  lng1 = p1(:,1); lat1 = p1(:,2);
  lng2 = p2(:,1); lat2 = p2(:,2);
  dLng = lng2 - lng1;
  dLat = lat2 - lat1;

  ## Equirectangular approximation
  c = norm(pi/180*[dLng .* cosd(lat1), dLat], 'rows');
  d = 6371000 * c;			# 6371000 is the mean Earth radius [m]

  ## Add floor penalty
  d += floorpenalty * abs(p1(:,3) - p2(:,3));
endfunction


## Distance given coordinates in their measurement unit
## Points P1 and P2 are [x y f] or nx3 matrices
function d = localdistance(p1, p2, floorpenalty = 0)
  xydiff = (p2-p1)(:,1:2);		# diffs of xy coordinates
  d = norm(xydiff, 'rows');

  ## Add floor penalty
  d += floorpenalty * abs(p1(:,3) - p2(:,3));
endfunction

#

## Compute path statistics
function create_path_stats (directory, distance)
  gtruthfile = [directory "/groundTruth.txt"];
  if !exist(gtruthfile, 'file')
    warning("%s does not exist: cannot create path statistics\n", gtruthfile);
  else
    [l x y f] = textread(gtruthfile, "%s%f%f%f");
    floorstart = 1 + [0; find(diff(f))];
    floorend = [find(diff(f)); length(f)];
    sumfloors = 0;
    sumpoints = 0;
    sumlen = 0;
    printf("\nFloor   |   horizontal length\n-----------------------------\n");
    for fidx = 1:length(floorstart);
      sidx = floorstart(fidx);
      eidx = floorend(fidx);
      spos = [x y f](sidx:eidx-1, :);
      epos = [x y f](sidx+1:eidx, :);
      points = eidx - sidx + 1;
      #wgs84distance(spos, epos)
      len = sum(feval(distance, spos, epos));
      printf("%d       |%3d points |%6.1f m\n",
	     f(floorstart(fidx)), points, len );
      sumfloors += 1;
      sumpoints += points;
      sumlen += len;
      if (eidx < length(x))	# if floor change ahead
	len = feval(distance, [x y f](eidx, :), [x y f](eidx+1, :));
	printf("%d -> %d  |           |%6.1f m\n",
	       f(eidx), f(eidx+1), len );
	sumlen += len;
      endif
    endfor
    printf("Totals -----------------------\n%d paths |%3d points |%6.1f m\n",
	   sumfloors, sumpoints, sumlen);
  endif

  buttonsfile = [directory "/buttonsPressed.log"];
  if exist(buttonsfile, 'file')
    [bpt bpl bpx bpy bpf] = textread(buttonsfile, "%f%s%f%f%f", 'delimiter', ":");
    sumtime = ceil(range(bpt) / 1000);
    printf("\nTotal time taken %d'%d\"",
	   fix(sumtime/60), rem(sumtime, 60));
    if exist("sumlen", 'var')
      printf(", average speed %.1f m/s", sumlen/sumtime);
    endif
    printf("\n");
  endif

  printf("\n");

endfunction


#

## Draw scenario graph
## Read the data files, put the reference coordinates in RX RY and the
## estimated coordinates in EX EY, draw the estimated path with the map image
## in background, if available
function [rl rt rx ry rf et ex ey ef] = scenariograph (s, directory, doplot, domovie)

  buttonsfile = [directory "/buttonsPressed.log"];
  positionsfile = [directory "/positions.log"];
  groundtruthfile = [directory "/groundTruth.txt"];

  ## Set up the map imagess
  alpha = 0.7;	    # transparency of map image
  light = 0.9;	    # light level of grey behind transparent map image

  if (doplot)
    for m = 1:numel(s.mapinfo)
      ## Create coordinate conversion matrix pixel_to_ref
      mapfile = [directory "/" s.mapinfo(m).map ".png"];
      cornersfile = [directory "/" s.mapinfo(m).map ".corners"];
      pw = imfinfo(mapfile).Width;
      ph = imfinfo(mapfile).Height;
      s.mapinfo(m).psz = [pw ph];
      corners_pixel = [0,0; pw,0; pw,-ph; 0,-ph];
      corners_ref = load('-ascii', cornersfile); # Ignore Octave headers
      pixel_to_ref = cp2tform(corners_pixel, corners_ref, 'projective');
      s.mapinfo(m).p2r = pixel_to_ref; # store it in s.mapinfo

      ## Draw the map as background
      mapimg = alpha*light*255 + (1-alpha)*flipud(fileimread(mapfile));
      ah = subplot(s.spsr, s.spsc, s.mapinfo(m).spi);
      if !isempty(s.mapinfo(m).spp)	# spp contains explicit position info
	## Scale the map and center it with respect to reference
	rm = s.mapinfo(m).rm;		# index of reference map
	if (s.mapinfo(m).rm != 0)	# scale the map
	  scale = s.mapinfo(m).psz ./ s.mapinfo(rm).psz; # [w h] scale
	  s.mapinfo(m).spp(3:4) .*= scale;
	  s.mapinfo(m).spp(1) -= s.mapinfo(m).spp(3) * (scale(1)-1)/2;
	endif
	set(ah, 'position', s.mapinfo(m).spp);
      endif
      s.mapinfo(m).ah = ah;			# axis handle
      imshow(mapimg);
      #keyboard
      axis('xy', 'image', 'off', 'nolabel'); # should use 'ij' and no flipud
      hold on
    endfor
  endif

  ## Read button pressed time, button label, reference coordinates (ground truth)
  [bpt bpl bpx bpy bpf] = textread(buttonsfile, "%f%s%f%f%f", 'delimiter', ":");

  assert(issorted(bpt), "non-increasing times in %s", buttonsfile);

  ## If the ground truth file is present, the bpx and bpy coordinates that
  ## we have just read are dummy ones, and we should recover them from the
  ## ground truth file; also check that the dummy values are 0
  assert(all([bpx; bpy; bpf] == 0),
	 "%s is present, but not all coordinates in %s are 0\n",
	 groundtruthfile, buttonsfile);
  [gtl gtx gty gtf] = textread(groundtruthfile, "%s%f%f%f");

  ## Check labels and floor
  gtl = strtrim(gtl); bpl = strtrim(bpl); # clean up labels
  assert(numel(gtl) == numel(unique(gtl)), "non-unique labels in %s", groundtruthfile);
  assert(numel(bpl) == numel(unique(bpl)), "non-unique labels in %s", buttonsfile);
  assert(all(round(gtf) == gtf), "non-integer floor in %s", groundtruthfile);

  ## Find indices of button pressures into ground truth by matching
  ## the button pressure labels bpl to ground truth labels gtl
  [found idx] = ismember(bpl, gtl);
  assert(all(found), "%s does not contain all labels in %s\n",
	 groundtruthfile, buttonsfile);
  assert(issorted(idx), "labels in %s not sorted as %s", buttonsfile, groundtruthfile);
  if (length(bpl) != length(gtl))
    warning("%d ground truth labels, but %d button pressures\n",
	    length(gtl), length(bpl));
  endif

  ## Set reference coordinates of button pressures
  bpx = gtx(idx); bpy = gty(idx); bpf = gtf(idx); # button pressures coordinates
  bpl = gtl(idx);				  # button pressure labels

  rl = bpl; rt = bpt;			# OUTPUT labels and time
  rx = bpx; ry = bpy; rf = bpf;		# OUTPUT reference coordinates

  ## Read estimation time, estimated coordinates (competitor's output)
  [pt px py pf] = textread(positionsfile, "%f%f%f%f");

  ## Check floor
  assert (all(round(pf) == pf), "non-integer floor in %s", positionsfile);

  ## Check that times are (almost) increasing and possibly correct them
  maxdecrease = 550;			# milliseconds
  minptdiff = min(diff(pt));
  if (minptdiff < 0)
    if (-minptdiff < maxdecrease)
      warning("non-increasing times in %s: max decrease %d ms\n",
	      positionsfile, -minptdiff);
      [pt idx] = sort(pt);
      px = px(idx); by = py(idx); pf = pf(idx);
    else
      error("non-increasing times in %s: max decrease %d > %d ms\n",
	    positionsfile, -minptdiff, maxdecrease);
    endif
  endif

  ## Check that first pos precedes first button pressure,
  ## possibly correct adding one pos at the beginning
  if (pt(1) > bpt(1))
    warning("%s starts after first button press, adding one initial log position\n",
	    positionsfile);
    pt = [bpt(1); pt];
    px = [px(1); px]; py = [py(1); py]; pf = [pf(1); pf];
  endif

  pidx = lookup(pt, bpt);			  # indices of competitor estimates
  bex = px(pidx); bey = py(pidx); bef = pf(pidx); # estimates at time of button pressed
  et = pt(pidx); ex = bex; ey = bey; ef = bef;	  # OUTPUT estimated coordinates
  assert(all(rt >= et), "internal consistency check failed");

  if (!doplot) return; endif;	# stop here after reading data files

  ## Draw the estimated path, estimated points, real points, error segments
  #########################################################################

  ## Create coordinates of error lines for plotting, NA is used to create
  ## separate segments when plotting
  errx = reshape([bpx ex NA(size(bpx))]', [], 1);
  erry = reshape([bpy ey NA(size(bpy))]', [], 1);
  wfidx = (bpf != bef);			# indices of bad floor estimates

  ## Do the plotting, for each img do the plot for all points in the
  ## corresponding floor
  if (domovie)
    printf("Creating movies:");
  else
    printf("Creating plots:");
  endif

  for m = 1:numel(s.mapinfo)
    printf([" " s.mapinfo(m).map]);	# show advancement
    fflush(1);				# necessary for Windows afer printf
    pixel_to_ref = s.mapinfo(m).p2r;	# tform used for coords conversion
    ph = s.mapinfo(m).psz(2);		# map height in pixels
    [ppx, ppy] = to_pixel_coords(ph, pixel_to_ref, px, py);
    [perrx, perry] = to_pixel_coords(ph, pixel_to_ref, errx, erry);
    [pbrx, pbry] = to_pixel_coords(ph, pixel_to_ref, bpx, bpy);
    [pbex, pbey] = to_pixel_coords(ph, pixel_to_ref, bex, bey);

    pepx = pbex;		 # pbex is estimated path, use ppx for whole path
    pepy = pbey;		 # pbey is estimated path, use ppy for whole path
    depx = diff(pepx); depy = diff(pepy); depf = diff(bpf);
    dlen = norm([depx depy], 'rows');	# used to set the arrow head sizes below

    ofidx = (rf != s.mapinfo(m).floor);	# index of other floor's points
    ofidx3 = repmat(ofidx', 3, 1);	# repeat each element three times
    pepx(ofidx) = pepy(ofidx) = NA;
    perrx(ofidx3) = perry(ofidx3) = NA;
    pbrx(ofidx) = pbry(ofidx) = NA;
    pbex(ofidx) = pbey(ofidx) = NA;

    msize = 2.5;			# size of markers
    lwidth = 2;				# width of thick lines
    hsize = 5;				# size of arrow heads

    ## Ground truth
    plot(s.mapinfo(m).ah,
    	 pbrx, pbry, "og;truth points;", 'markersize', msize, 'markerfacecolor', 'green');
    axis('equal');

    hold on
    if (domovie)

      moviefn = [s.mapinfo(m).map ".mp4"];
      movie('init', moviefn)
      plot(s.mapinfo(m).ah,		# first frame
	   perrx(1:2), perry(1:2),
	   pbex(1), pbey(1), "ob;estimated points;", 'markersize', msize);
      for f = 1:length(pepx)-1		# next frames
	plot(s.mapinfo(m).ah,
	     perrx(3*f+(0:1)), perry(3*f+(0:1)), "-r;errors;",
	     pbex(f+1), pbey(f+1), "ob;estimated points;", 'markersize', msize,
	     pbex(wfidx), pbey(wfidx), 'markerfacecolor', 'red', 'markersize', msize);
	if (dlen(f) > 0)
	  arrowcolour = 'blue';
	  if depf(f)
	    arrowcolour = 'cyan';
	  endif
	  quiver(pepx(f), pepy(f), depx(f), depy(f), 0, 'color', arrowcolour,
  		 'linewidth', lwidth, 'maxheadsize', hsize./dlen(f));
	endif
	legend('off');
	movie('add', moviefn);
      endfor
      movie('close', moviefn)

    else

      plot(s.mapinfo(m).ah,
	   pepx, pepy, "-b;estimated path;", 'linewidth', lwidth,
	   perrx, perry, "-r;errors;",
	   pbex, pbey, "ob;estimated points;", 'markersize', msize,
	   pbex(wfidx), pbey(wfidx), "or;wrong floor;", 'markerfacecolor', 'red', 'markersize', msize);

      ## Overwrite the estimated path plotted above with arrows.
      ## The head size is set for each arrow using the "hsize" variable.
      ## The plot instruction above must nonetheless include
      ## the estimated path so that it appears in the legend.
      for i = 1:length(pepx)-1
	if (dlen(i) > 0)
	  arrowcolour = 'blue';
	  if depf(i)
	    arrowcolour = 'cyan';
	  endif
	  quiver(pepx(i), pepy(i), depx(i), depy(i), 0, 'color', arrowcolour,
  		 'linewidth', lwidth, 'maxheadsize', hsize./dlen(i));
	endif
      endfor

      ## Plot accessory elements
      th = title(s.mapinfo(m).ah, s.mapinfo(m).title);
      set(th, 'fontname', 'arial', 'fontweight', 'light');
      if (m == s.spi_legax)
	hleg = legend('location', 'northeastoutside');
	legpos = legpos = get(hleg, 'position');
	## Setting the legend position does not work well with gnuplot
	set(hleg, 'position', legpos+[0 -0.15 0 0]); legend hide; legend show;
      else
	legend off;
      endif
    endif

    hold off

  endfor

  printf(" done.\n");

endfunction


#
################################################################
## Main program
################################################################

## Set default values
exist("basedir") || (basedir = ".");
exist("doplot") || (doplot = true);
exist("dopath") || (dopath = true);
exist("domovie") || (domovie = false);
(domovie) && (doplot = true);
if (! exist("dirname"))
  dirname = chosendir(basedir);
endif
assert(!isempty(dirname), "no directory chosen, exiting\n");

directory = [basedir "/" dirname];

distancefile = [directory "/distance"];
legendfile = [directory "/legend.png"];
errorfile = [directory "/error.report"];
qgisfile = [directory "/qgis_errors.txt"];

## Get the distance type
distance = textread(distancefile, "%s"){1};

## Print statistics
create_path_stats(directory, distance);

## Select scenario
for [s name] = scenarios
  notfound = false;
  for mapfile = {s.mapinfo.map}
    if !exist([directory "/" mapfile{} ".png"], 'file')
      notfound = true;
      break;
    endif
  endfor
  if (notfound)
    continue;					# check next scenario
  else
    printf(["Using scenario '" name "'\n"]);
    scnr = s;
    break
  endif
endfor
if (notfound)
  printf(["Using no scenario\n"]);
  scnr = struct("floorpenalty", 0, "reporttitle", "");
  if (doplot)
    printf("No known scenario in the selected directory: setting doplot=false\n");
    doplot = false;
  endif
endif
fflush(1);			# necessary for Windows

if (doplot)
  if (! strncmp(version, "4.0", 3))
    warning("cannot plot: need Octave 4.0.  Generating report only.\n");
    doplot = false;
  elseif isempty(pkg('list', 'image'))
    warning("cannot plot: package 'image' not installed.  Generating report only.\n");
    doplot = false
  else
    pkg load image				# for tform functions
    warning('off', 'Octave:GraphicsMagic-Quantum-Depth');

    if isempty(pkg('list', 'plot'))
      warning("package 'plot' not installed.  Install it for best plot quality.");
      plotpkg = false;
    else
      pkg load plot				# for tics()
      plotpkg = true;
    endif

    graphics_toolkit gnuplot			# needed in Octave 4.0

    ## Prevent plotting, do that at the end only
    figure(1, 'visible', 'off');
    clf(1);
  endif
endif

if (doplot)
  set(gcf(),					# format for landscape on A4 paper
      'papertype', 'a4', 'paperorientation', 'landscape',
      'paperunits', 'centimeters', 'paperposition', [1 1.5 19 27]);

  ## Generate path file
  if (dopath) draw_path (scnr, directory); endif
endif

## Read data, possibly draw the path with background maps,
## return the reference coordinates and the estimated coordinates
[rl rt rx ry rf et ex ey ef] = scenariograph(scnr, directory, doplot, domovie);

## Compute error statistics
poserr = feval(distance, [rx ry rf], [ex ey ef], scnr.floorpenalty);
ninetiethperc = quantile(poserr, 90/100, 1, 7)	# ninetieth percentile of error
thirdq = quantile(poserr, 3/4, 1, 7)		# third quartile of error

## Generate score file
if (doplot)

  lah = subplot(scnr.spsr, scnr.spsc, scnr.spi_legend);
  axis(lah, 'off');

  ## Legend
  if (exist(legendfile, 'file') && scnr.spi_legax == 0)
    isempty(scnr.spp_legend) || set(lah, 'position', scnr.spp_legend);
    mapimg = flipud(fileimread(legendfile));
    imshow(mapimg);
    axis('xy', 'image', 'off', 'nolabel');	# should use 'ij' and no flipud
  endif

  ## Global title
  ## Does not work with Octave 4.0 and gnuplot, so we make do with title()
  #gth = annotation('textbox', [0.4 0.95 0.2 0.2], 'string', dirname);
  gth = title(lah, dirname);
  set(gth, 'interpreter', 'none');

  ## Distribution graph
  ah = subplot(scnr.spsr, scnr.spsc, scnr.spi_distrib);
  isempty(scnr.spp_distrib) || set(ah, 'position', scnr.spp_distrib);
  maxerr = 20;				# max error to plot
  errno = length(poserr);			# number of error
  stairs([0; sort(poserr)], (0:errno)/errno);
  title("distribution");
  axis([0 maxerr 0 1], 'square');
  if (plotpkg)				# if plot package loaded
    tics('y', 0:0.25:1);
  endif;
  grid on
  hold on
  plot(thirdq*[1 1], [0 1], 'linewidth', 2, 'color', 'red');

  ## Statistics
  if compare_versions(version, "4.0.3", '<')
    s = sprintf("median: %.1f\nmean: %.1f\nrms: %.1f\n\n3^{rd} quartile: %.1f",
    		median(poserr), mean(poserr), std(poserr, 1), thirdq);
    text(scnr.statx, scnr.staty, s);
  else
    xp = scnr.statx; yp = scnr.staty;		# x, y position
    il = 0.13;				# interline space
    ln = 0;					# line number
    text(xp, yp, "Error stats [m]", 'fontsize', 12);
    text(xp, yp-il*(ln+=2), sprintf("median: %.1f", median(poserr)));
    text(xp, yp-il*(ln+=1), sprintf("mean: %.1f", mean(poserr)));
    text(xp, yp-il*(ln+=1), sprintf("rms: %.1f", std(poserr, 1)));
    text(xp, yp-il*(ln+=2), sprintf("90^{th} perc.: %.1f", ninetiethperc));
    text(xp, yp-il*(ln+=2), sprintf("3^{rd} quartile: %.1f", thirdq),
	 'fontsize', 12, 'fontweight', 'bold');
  endif

  ## Generate score file
  scorefile = [directory "/score.pdf"];
  print(scorefile);

endif

## Print Qgis data
fid = fopen(qgisfile, 'w');
    fprintf(fid,"id;x;y;z;id1;x1;y1;z1\n");
for l = 1:length(rl)
    fprintf(fid,"%s;",rl{l});
    fprintf(fid,"%.6f;%6f;%d;", [rx ry rf](l,:));
    fprintf(fid,"%s",rl{l});
    if (rf(l) != ef(l)) fprintf(fid,"*") endif
    fprintf(fid,";%.6f;%6f;%d\n", [ex ey ef](l,:));
endfor
fclose(fid);

## Print error report
fid = fopen(errorfile, 'w');
fprintf(fid, [scnr.reporttitle "\n\n"]);
fprintf(fid, "\t\t\t================ %s ================\n\n", dirname);
fprintf(fid, "\t\t\t  === Third quartile of error: %.1f [m] ===\n\n", thirdq);
fprintf(fid, "Floor penalty is %.1f metres\n", scnr.floorpenalty);
fprintf(fid, "Error statistics [m]: median: %.1f  mean: %.1f  RMS: %.1f\n\n",
	median(poserr), mean(poserr), std(poserr, 1));
fprintf(fid, "\
Label\tReference coordinates\t\tEstimated coordinates\tError[m]  Error[s]\
\n\n");
for l = 1:length(rl)
  fprintf(fid, "%5s%14.6f%10.6f%3d\t%11.6f%10.6f%3d",
	  rl{l}, [rx ry rf ex ey ef](l,:));
  if (rf(l) != ef(l))
    fprintf(fid, "%+8.1f", poserr(l));
  else
    fprintf(fid, "%8.1f", poserr(l));
  endif
  fprintf(fid, "%10.1f\n", (rt(l)-et(l))/1000);
endfor
fclose(fid);

## Set showplot to false in production
showplot = false;
if (doplot)
  if (showplot)
    figure(1, 'visible', 'on');
  else
    close(1);
  endif
endif

## Cleanup
clear doplot domovie basedir dirname
